// -*- c++ -*-
/*
 * Copyright (C) 2011 Hans Vandierendonck (hvandierendonck@acm.org)
 * Copyright (C) 2011 George Tzenakis (tzenakis@ics.forth.gr)
 * Copyright (C) 2011 Dimitrios S. Nikolopoulos (dsn@ics.forth.gr)
 * 
 * This file is part of Swan.
 * 
 * Swan is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * Swan is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with Swan.  If not, see <http://www.gnu.org/licenses/>.
 */

/* ltickets.h
 * This file implements a ticket-based task graph where pending tasks
 * are kept on lists as in ecltaskgraph.
 * This is not working correctly yet.
 */
#ifndef LTICKETS_H
#define LTICKETS_H

#include <cstdint>
#include <iostream>

#include "swan/swan_config.h"
#include "swan/wf_frames.h"
#include "swan/lfllist.h"
#include "swan/lock.h"
#include "swan/functor/tkt_ready.h"
#include "swan/taskgraph/ready_list_tg.h"

namespace obj {

// The type holding the depth of an object in the task graph.
// The depth should not wrap around. If so, adopted implementations
// of max should be used in several dependency action traits.
typedef uint64_t depth_t;

// ----------------------------------------------------------------------
// link_metadata: task graph metadata per stored frame
// ----------------------------------------------------------------------
class pending_metadata;

class link_metadata {
    link_metadata * prev, * next; // no need to initialize
    friend class dl_list_traits<link_metadata>;
};

//----------------------------------------------------------------------
// Traits for accessing elements stored in a dl_list<>, in global
// namespace.
//----------------------------------------------------------------------
} // end namespace obj because the traits class must be in global namespace

template<>
struct dl_list_traits<obj::link_metadata> {
    typedef obj::link_metadata T;
    typedef obj::link_metadata ValueType;

    // not implemented -- static size_t get_depth( T * elm );
    // not implemented -- static bool is_ready( T * elm );

    static void set_prev( T * elm, T * prev ) { elm->prev = prev; }
    static T * get_prev( T const * elm ) { return elm->prev; }
    static void set_next( T * elm, T * next ) { elm->next = next; }
    static T * get_next( T const * elm ) { return elm->next; }

    static ValueType * get_value( T const * elm ) {
	return const_cast<ValueType *>( elm ); // cheat
    }
};

namespace obj { // reopen

// ----------------------------------------------------------------------
// taskgraph: task graph roots in ready_list
// ----------------------------------------------------------------------
class pending_metadata;

typedef ::taskgraph<pending_metadata, link_metadata,
		    dl_head_list<link_metadata> > taskgraph;

// ----------------------------------------------------------------------
// list_tags: tag storage for all dependency usage types
// ----------------------------------------------------------------------
class task_metadata;
class ltkt_metadata;

// list_tags stores the accessed generation and links the task on the task list.
// Required for all types
class list_tags {
    typedef uint32_t tag_t; // links to ltkt_metadata::tag_t

    tag_t rd_tag;
    tag_t wr_tag;
#if OBJECT_COMMUTATIVITY
    tag_t c_tag;
#endif
#if OBJECT_REDUCTION
    tag_t r_tag;
#endif

    list_tags * it_next;
    task_metadata * st_task;

    friend class serial_dep_traits;
    template<typename MetaData, typename Task, template<typename U> class DepTy>
    friend class dep_traits;
    friend class ltkt_metadata;
    friend class sl_list_traits<list_tags>;

public:
    inline bool arg_is_ready( ltkt_metadata * ltkt );
    inline bool is_current_generation( ltkt_metadata * ltkt );

    inline void copy_tags( ltkt_metadata * ltkt );
};

} // end namespace obj because the traits class must be in global namespace

template<>
struct sl_list_traits<obj::list_tags> {
    typedef obj::list_tags T;
    typedef obj::task_metadata ValueType;

    static void set_next( T * elm, T * next ) { elm->it_next = next; }
    static T * get_next( T const * elm ) { return elm->it_next; }

    static ValueType * get_value( T const * elm ) { return elm->st_task; }
};

namespace obj { // reopen

// ----------------------------------------------------------------------
// ltkt_metadata: dependency-tracking metadata (not versioning)
// ----------------------------------------------------------------------
class ltkt_metadata {
public:
    struct fifo_like {
	typedef uint32_t ctr_t;
	ctr_t head;
	ctr_t tail;
    public:
	fifo_like() : head( 0 ), tail( 0 ) {
	    assert( (intptr_t(&head) & (sizeof(head)-1)) == 0
		    && "Alignment of head field not respected" );
	    assert( (intptr_t(&tail) & (sizeof(tail)-1)) == 0
		    && "Alignment of tail field not respected" );
	}

	ctr_t adv_head() { return __sync_fetch_and_add( &head, 1 ); }
	// Note: tail increment is covered by parent lock only if we use
	// hyperobjects intra-procedurally. If we use them "wrongly" then
	// the tail update must be atomic.
	// ctr_t adv_tail() { return tail++; } // covered by parent lock
	ctr_t adv_tail() { return __sync_fetch_and_add( &tail, 1 ); }
	bool empty() const volatile { return head == tail; }
	bool chk_tag( ctr_t tag ) const volatile { return head == tag; }
	ctr_t get_tag() const { return tail; }

	friend std::ostream &
	operator << ( std::ostream & os, const fifo_like & f );
    };
    typedef fifo_like::ctr_t tag_t;

private:
    fifo_like writers;            // head and tail counter for writers
    fifo_like readers;            // head and tail counter for readers
#if OBJECT_COMMUTATIVITY
    fifo_like commutative;        // head and tail counter for commutative IO
    cas_mutex mutex;              // ensure exclusion on commutative operations
#endif
#if OBJECT_REDUCTION
    fifo_like reductions;         // head and tail counter for readers
#endif
    depth_t depth;                // depth in task graph

    sl_list<list_tags> tasks;     // all pending tasks not on ready list
    cas_mutex tmutex;

public:
    ltkt_metadata() : depth( 0 ) { }
    ~ltkt_metadata() {
	assert( readers.empty()
		&& "Must have zero readers when destructing obj_version" );
	assert( writers.empty()
		&& "Must have zero writers when destructing obj_version" );
    }

    // External interface
    bool rename_is_active() const volatile {
        return rename_has_readers() || rename_has_writers();
    }
    bool rename_has_readers() const volatile { return has_readers(); }
    bool rename_has_writers() const volatile {
	return has_writers()
#if OBJECT_COMMUTATIVITY
	    | has_commutative()
#endif
#if OBJECT_REDUCTION
	    | has_reductions()
#endif
	    ;
    }

    // Track oustanding readers with a head and tail counter
    void add_reader() { readers.adv_tail(); }
    void del_reader() { readers.adv_head(); }
    bool chk_reader_tag( tag_t w ) const volatile { return readers.chk_tag(w); }
    bool has_readers() const volatile { return !readers.empty(); }
    tag_t get_reader_tag() const { return readers.get_tag(); }

    // Track outstanding writers with a head and tail counter, both adding up.
    // This organization allows us to easily implement inout dependencies where
    // the object with an inout dependency is never renamed. Each successive
    // inout dependency gets a different value of the writers variable, a "tag"
    // that must be matched for waking up pending children.
    void add_writer() { writers.adv_tail(); } 
    void del_writer() { writers.adv_head(); }
    bool chk_writer_tag( tag_t w ) const volatile { return writers.chk_tag(w); }
    bool has_writers() const volatile { return !writers.empty(); }
    tag_t get_writer_tag() const { return writers.get_tag(); }

#if OBJECT_COMMUTATIVITY
    // Track commutative IO
    void add_commutative() { commutative.adv_tail(); }
    void del_commutative() { commutative.adv_head(); }
    bool chk_commutative_tag( tag_t w ) const volatile {
	return commutative.chk_tag(w);
    }
    bool has_commutative() const volatile { return !commutative.empty(); }
    tag_t get_commutative_tag() const { return commutative.get_tag(); }

    // There is no lock operation - because there is no reason to wait...
    bool commutative_try_acquire() { return mutex.try_lock(); }
    void commutative_release() { mutex.unlock(); }
#endif

#if OBJECT_REDUCTION
    // Track reductions
    void add_reduction() { reductions.adv_tail(); }
    void del_reduction() { reductions.adv_head(); }
    bool chk_reduction_tag( tag_t w ) const volatile {
	return reductions.chk_tag(w);
    }
    bool has_reductions() const volatile { return !reductions.empty(); }
    tag_t get_reduction_tag() const { return reductions.get_tag(); }
#endif

    depth_t get_depth() const { return depth; }
    void update_depth( depth_t d ) {
	// In principle, an outdep has depth == 0 and an inoutdep has depth <= d
	// because d = max(depth_in) over all indep and inoutdep.
	// Note: an unversioned outdep has depth > 0 and must also be
	// included in the depth computation of the frame!
	// depth = std::max( depth, d );
	assert( depth <= d );
	depth = d;
    }

#ifdef UBENCH_HOOKS
    void update_depth_ubench( depth_t d ) { depth = d; }
#endif

    inline void add_task( task_metadata * t, list_tags * tags );
    inline void wakeup( taskgraph * tg );

    friend std::ostream & operator << ( std::ostream & os, const ltkt_metadata & md );
};

void list_tags::copy_tags( ltkt_metadata * md ) {
    rd_tag  = md->get_reader_tag();
    wr_tag  = md->get_writer_tag();
#if OBJECT_COMMUTATIVITY
    c_tag  = md->get_commutative_tag();
#endif
#if OBJECT_REDUCTION
    r_tag  = md->get_reduction_tag();
#endif
}

bool list_tags::arg_is_ready( ltkt_metadata * md ) {
    // TODO: depends on annotation! - else everything serialized!
    bool r = md->chk_reader_tag( rd_tag )
#if OBJECT_COMMUTATIVITY
	& md->chk_commutative_tag( c_tag )
#endif
#if OBJECT_REDUCTION
	& md->chk_reduction_tag( r_tag )
#endif
	& md->chk_writer_tag( wr_tag );
    errs() << "arg is ready on md=" << *md
	   << " with rd=" << rd_tag
	   << " with c=" << c_tag
	   << " with r=" << r_tag
	   << " with wr=" << wr_tag
	   << ": " << r << std::endl;
    return r;
}

bool list_tags::is_current_generation( ltkt_metadata * md ) {
    size_t num_checked = 4;
    if( md->chk_reader_tag( rd_tag ) )
	--num_checked;
    if( md->chk_writer_tag( rd_tag ) )
	--num_checked;
#if OBJECT_COMMUTATIVITY
    if( md->chk_commutative_tag( c_tag ) )
	--num_checked;
#else
    --num_checked;
#endif
#if OBJECT_REDUCTION
    if( md->chk_reduction_tag( r_tag ) )
	--num_checked;
#else
    --num_checked;
#endif
    return num_checked <= 1; // 3 tags equal, one differs.
}

class token_metadata {
public:
    struct fifo_like {
	typedef uint32_t ctr_t;
	ctr_t head;
	ctr_t tail;
    public:
	fifo_like() : head( 0 ), tail( 0 ) {
	    assert( (intptr_t(&head) & (sizeof(head)-1)) == 0
		    && "Alignment of head field not respected" );
	    assert( (intptr_t(&tail) & (sizeof(tail)-1)) == 0
		    && "Alignment of tail field not respected" );
	}

	ctr_t adv_head() { return __sync_fetch_and_add( &head, 1 ); }
	// Note: tail increment is covered by parent lock only if we use
	// hyperobjects intra-procedurally. If we use them "wrongly" then
	// the tail update must be atomic.
	// ctr_t adv_tail() { return tail++; } // covered by parent lock
	ctr_t adv_tail() { return __sync_fetch_and_add( &tail, 1 ); }
	bool empty() const volatile { return head == tail; }
	bool chk_tag( ctr_t tag ) const volatile { return head == tag; }
	ctr_t get_tag() const { return tail; }

	friend std::ostream &
	operator << ( std::ostream & os, const fifo_like & f );
    };
    typedef fifo_like::ctr_t tag_t;

private:
    fifo_like writers;            // head and tail counter for writers
    fifo_like readers;            // head and tail counter for readers
    depth_t depth;                // depth in task graph

public:
    token_metadata() : depth( 0 ) { }
    ~token_metadata() {
	assert( readers.empty()
		&& "Must have zero readers when destructing obj_version" );
	assert( writers.empty()
		&& "Must have zero writers when destructing obj_version" );
    }

    // External interface - never do renaming
    bool rename_is_active() const volatile { return true; }
    bool rename_has_readers() const volatile { return true; }
    bool rename_has_writers() const volatile { return true; }

    // Track oustanding readers with a head and tail counter
    void add_reader() { readers.adv_tail(); }
    void del_reader() { readers.adv_head(); }
    bool chk_reader_tag( tag_t w ) const volatile { return readers.chk_tag(w); }
    bool has_readers() const volatile { return !readers.empty(); }
    tag_t get_reader_tag() const { return readers.get_tag(); }

    // Track outstanding writers with a head and tail counter, both adding up.
    // This organization allows us to easily implement inout dependencies where
    // the object with an inout dependency is never renamed. Each successive
    // inout dependency gets a different value of the writers variable, a "tag"
    // that must be matched for waking up pending children.
    void add_writer() { writers.adv_tail(); } 
    void del_writer() { writers.adv_head(); }
    bool chk_writer_tag( tag_t w ) const volatile { return writers.chk_tag(w); }
    bool has_writers() const volatile { return !writers.empty(); }
    tag_t get_writer_tag() const { return writers.get_tag(); }

    depth_t get_depth() const { return depth; }
    void update_depth( depth_t d ) {
	// In principle, an outdep has depth == 0 and an inoutdep has depth <= d
	// because d = max(depth_in) over all indep and inoutdep.
	// Note: an unversioned outdep has depth > 0 and must also be
	// included in the depth computation of the frame!
	// depth = std::max( depth, d );
	assert( depth <= d );
	depth = d;
    }
};

// Some debugging support
inline std::ostream &
operator << ( std::ostream & os, const ltkt_metadata::fifo_like & f ) {
    return os << '{' << f.head << ", " << f.tail << '}';
}

inline std::ostream & operator << ( std::ostream & os, const ltkt_metadata & md ) {
    os << "lticket_md={" << &md << ", readers=" << md.readers
       << ", writers=" << md.writers
#if OBJECT_COMMUTATIVITY
       << ", commutative=" << md.commutative
#endif
#if OBJECT_REDUCTION
       << ", reductions=" << md.reductions
#endif
       << '}';
    return os;
}

// ----------------------------------------------------------------------
// A function to do the first scan over the arguments in the computation
// of the depth of the newly created task and its arguments.
// The first scan updates the depth of the frame based on its in and inout
// arguments.
// The second scan is part of dep_traits<>::arg_issue() and updates the
// depth of the out and inout arguments.
// ----------------------------------------------------------------------
// Update-depth functor
template<typename MetaData, typename Task>
struct update_depth_functor {
    Task * fr;
    update_depth_functor( Task * fr_ ) : fr( fr_ ) { fr->reset_depth(); }

    template<typename T, template<typename U> class DepTy >
    inline
    bool operator () ( DepTy<T> & obj, typename DepTy<T>::dep_tags & sa ) {
	// Don't update depth for outdep dependencies
	if( !is_outdep< DepTy<T> >::value )
	    fr->update_depth( obj.get_version()->get_metadata()->get_depth() );
	return true;
    }
    template<typename T>
    inline
    bool operator () ( truedep<T> & obj, typename truedep<T>::dep_tags & sa ) {
	return true;
    }
};

#if STORED_ANNOTATIONS
// A "update depth" function to adjust the depth of the objects in the taskgraph
template<typename Task, typename MetaData, typename... Tn>
static inline void arg_update_depth_fn( Task * fr, task_data_t & td ) {
    update_depth_functor<MetaData, Task> dfn( fr );
    arg_apply_stored_fn<update_depth_functor<MetaData, Task>,Tn...>( dfn, td );
}
#else
// A "update depth" function to adjust the depth of the objects in the taskgraph
template<typename Task, typename MetaData, typename... Tn>
static inline void arg_update_depth_fn( Task * fr, task_data_t & td ) {
    update_depth_functor<MetaData, Task> dfn( fr );
    arg_apply_fn<update_depth_functor<MetaData, Task>,Tn...>(
	dfn, td.get_args_ptr(), td.get_tags_ptr() );
}
#endif

// ----------------------------------------------------------------------
// task_metadata: dependency-tracking metadata for tasks (pending and stack)
// ----------------------------------------------------------------------
class full_metadata;

class task_metadata : public task_data_t {
    size_t incoming_count;
    depth_t depth;
    taskgraph * graph;

protected:
    // Default constructor
    task_metadata() : incoming_count( 0 ), graph( 0 ) {
	// errs() << "task " << this << " created" << std::endl;
 }
    	// ~task_metadata() { errs() << "task " << this << " finished" << std::endl; };
    // Constructor for creating stack frame from pending frame
    inline void create_from_pending( task_metadata * fr, full_metadata * );
    inline void convert_to_full( full_metadata * );

public:
    taskgraph * get_graph() { return graph; }

    template<typename... Tn>
    inline void create( full_metadata * ff );

    void reset_depth() { depth = 0; }
    depth_t get_depth() const { return depth; }
    void update_depth( depth_t d ) {
	depth = std::max( depth, d+1 );
    }

    // Ready counter
    void add_incoming() {
		// errs() << this << " add_incoming was " << incoming_count << std::endl;
	__sync_fetch_and_add( &incoming_count, 1 );
    }
    bool del_incoming() {
		// errs() << this << " del_incoming was " << incoming_count << std::endl;
	return __sync_fetch_and_add( &incoming_count, -1 ) == 1;
    }

    // Stubs required for the taskgraph variant, here meaningless.
    template<typename... Tn>
    void start_registration() {
#if STORED_ANNOTATIONS
	arg_update_depth_fn<task_metadata, ltkt_metadata>( this, get_task_data() );
#else
	arg_update_depth_fn<task_metadata, ltkt_metadata, Tn...>(
	    this, get_task_data() );
#endif
	add_incoming();
    }
    inline void stop_registration( bool wakeup = false );

    void start_deregistration() { }
    inline void stop_deregistration( full_metadata * parent );
};

// ----------------------------------------------------------------------
// pending_metadata: task graph metadata per pending frame
// ----------------------------------------------------------------------
class pending_metadata : public task_metadata, public link_metadata {
#if !STORED_ANNOTATIONS
    typedef bool (*ready_fn_t)( const task_data_t & );

    ready_fn_t ready_fn;
#endif

protected:
    pending_metadata()
#if !STORED_ANNOTATIONS
	: ready_fn( 0 )
#endif
	{ }

public:
    template<typename... Tn>
    void create( full_metadata * ff ) {
#if !STORED_ANNOTATIONS
	ready_fn = &arg_ready_fn<ltkt_metadata, task_metadata, Tn...>;
#endif
	task_metadata::create<Tn...>( ff );
    }

    bool is_ready() const {
#if PROFILE_WORKER
	extern __thread size_t num_tkt_evals;
	++num_tkt_evals;
#endif // PROFILE_WORKER
#if STORED_ANNOTATIONS
	return arg_ready_fn<ltkt_metadata, task_metadata>( get_task_data() );
#else
	return ready_fn && (*ready_fn)( get_task_data() );
#endif
    }

    static obj::link_metadata *
    get_link_md( pending_metadata * fr ) {
	return static_cast<obj::link_metadata *>( fr );
    }
    static const obj::link_metadata *
    get_link_md( const pending_metadata * fr ) {
	return static_cast<const obj::link_metadata *>( fr );
    }
    static obj::task_metadata *
    get_task_md( pending_metadata * fr ) {
	return static_cast<obj::task_metadata *>( fr );
    }
    static const obj::task_metadata *
    get_task_md( const pending_metadata * fr ) {
	return static_cast<const obj::task_metadata *>( fr );
    }

    static inline
    pending_metadata * get_from_task( task_metadata * task ) {
	return static_cast<pending_metadata *>( task );
    }
    static inline
    pending_metadata * get_from_link( link_metadata * link ) {
	return static_cast<pending_metadata *>( link );
    }
};

#if 0
//----------------------------------------------------------------------
// Traits for accessing elements stored in a dl_list<>, in global
// namespace.
//----------------------------------------------------------------------
class task_metadata;
} // end namespace obj because the traits class must be in global namespace

template<>
struct dl_list_traits<obj::pending_metadata> {
    typedef obj::pending_metadata T;
    typedef obj::pending_metadata ValueType;

    static size_t get_depth( T const * elm ) { return 0; } // TF( elm )->get_depth(); }
    static bool is_ready( T const * elm ) { return elm->is_ready(); }

    static void set_prev( T * elm, T * prev ) { LF( elm )->prev = prev; }
    static T * get_prev( T const * elm ) { return LF( elm )->prev; }
    static void set_next( T * elm, T * next ) { LF( elm )->next = next; }
    static T * get_next( T const * elm ) { return LF( elm )->next; }

    // static ValueType * get_value( const T * elm ); -- not implemented

private:
    static obj::link_metadata * LF( T * fr ) { return T::get_link_md( fr ); }
    static const obj::link_metadata * LF( T const * fr ) { return T::get_link_md( fr ); }
    static obj::task_metadata * TF( T * fr ) { return T::get_task_md( fr ); }
    static const obj::task_metadata * TF( T const * fr ) { return T::get_task_md( fr ); }
};

namespace obj { // reopen
#endif

// ----------------------------------------------------------------------
// taskgraph: task graph roots in ready_list
// ----------------------------------------------------------------------
} // end namespace obj

template<>
struct taskgraph_traits<obj::pending_metadata, obj::link_metadata, dl_head_list<obj::link_metadata> > {
    typedef obj::pending_metadata task_type;
    typedef obj::link_metadata stored_task_type;

    static obj::pending_metadata * get_task( obj::link_metadata * lnk ) {
	return obj::pending_metadata::get_from_link( lnk );
    }

    static bool acquire( obj::pending_metadata * pnd ) {
	return true; // TODO: pnd->acquire();
    }
};

namespace obj { // reopen

// ----------------------------------------------------------------------
// full_metadata: task graph metadata per full frame
// ----------------------------------------------------------------------
class full_metadata {
protected:
    taskgraph graph;

protected:
    full_metadata() { }

public:
    void push_pending( pending_metadata * frame ) { }

    pending_metadata * get_ready_task() {
	pending_metadata * t = graph.get_ready_task();
		// errs() << "get ready: " << t << std::endl;
	return t;
    }

    pending_metadata * get_ready_task_after( task_metadata * ) {
	return get_ready_task();
    }

    taskgraph * get_graph() { return &graph; }
};

// ----------------------------------------------------------------------
// Member function implementations with cyclic dependencies on class
// definitions.
// ----------------------------------------------------------------------
void
task_metadata::create_from_pending( task_metadata * from, full_metadata * ff ) {
    graph = ff->get_graph();
    depth = from->depth;
}

void
task_metadata::convert_to_full( full_metadata * ff ) {
    graph = ff->get_graph();
    assert( graph != 0 && "create_from_pending with null graph" );
}

template<typename... Tn>
void
task_metadata::create( full_metadata * ff ) {
    graph = ff->get_graph();
#if !STORED_ANNOTATIONS
    // acquire_fn = &arg_acquire_fn<ecltg_metadata,Tn...>;
    // errs() << "set graph in " << this << " to " << graph << "\n";
#endif
}

void task_metadata::stop_registration( bool wakeup ) {
    if( del_incoming() && wakeup )
	graph->add_ready_task( pending_metadata::get_from_task( this ) );
}

void task_metadata::stop_deregistration( full_metadata * parent ) { }

#define PQRTL(x) ((x)!=0?(x)->st_task:((task_metadata*)0))

void ltkt_metadata::
add_task( task_metadata * t, list_tags * tags ) {
    tags->st_task = t;

    tmutex.lock();
    if( !tags->arg_is_ready( this ) ) {
	t->add_incoming();
	tasks.push_back( tags );
		// errs() << "task list " << this << ": head=" << PQRTL(tasks.front()) << " tail=" << PQRTL(tasks.back()) << std::endl;
    }
    tmutex.unlock();
}

void ltkt_metadata::wakeup( taskgraph * graph ) {
    tmutex.lock();

    	// errs() << "wakeup of md=" << this << ": " << *this << std::endl;

    if( tasks.empty() ) {
	tmutex.unlock();
	return;
    }

    	// errs() << "task list " << this << ": head=" << PQRTL(tasks.front()) << " tail=" << PQRTL(tasks.back()) << std::endl;

    list_tags * head = tasks.front();
    list_tags * i_next = head;
    for( list_tags * i=head; i; i=i_next ) {
	if( !i->is_current_generation( this ) ) // based on tail counters
	    break;
	task_metadata * t = i->st_task;
	i_next = i->it_next;
	if( t->del_incoming() ) {
	    	// errs() << "task " << t << " ready" << std::endl;
	    graph->add_ready_task( pending_metadata::get_from_task( t ) );
	}
    }

    if( i_next )
	tasks.fastforward( i_next );
    else
	tasks.clear();

    	// errs() << "task list " << this << ": head=" << PQRTL(tasks.front()) << " tail=" << PQRTL(tasks.back()) << std::endl;

    tmutex.unlock();
}


// ----------------------------------------------------------------------
// Generic fully-serial dependency handling traits
// ----------------------------------------------------------------------

// A fully serialized version
class serial_dep_tags : public list_tags {
    friend class serial_dep_traits;
};

struct serial_dep_traits {
    static
    void arg_issue( task_metadata * fr,
		    obj_instance<ltkt_metadata> & obj,
		    serial_dep_tags * tags ) {
	ltkt_metadata * md = obj.get_version()->get_metadata();
	tags->copy_tags( md );
	md->add_writer();
	md->add_task( fr, tags );
	md->update_depth( fr->get_depth() );
    }
    static
    bool arg_ready( obj_instance<ltkt_metadata> obj, serial_dep_tags & tags ) {
	ltkt_metadata * md = obj.get_version()->get_metadata();
	return md->chk_reader_tag( tags.rd_tag )
#if OBJECT_COMMUTATIVITY
	    & md->chk_commutative_tag( tags.c_tag )
#endif
#if OBJECT_REDUCTION
	    & md->chk_reduction_tag( tags.r_tag )
#endif
	    & md->chk_writer_tag( tags.wr_tag );
    }
    static
    bool arg_ini_ready( const obj_instance<ltkt_metadata> obj ) {
	const ltkt_metadata * md = obj.get_version()->get_metadata();
	return !md->has_readers()
#if OBJECT_COMMUTATIVITY
	    & !md->has_commutative()
#endif
#if OBJECT_REDUCTION
	    & !md->has_reductions()
#endif
	    & !md->has_writers();
    }
    static
    void arg_release( task_metadata * fr, obj_instance<ltkt_metadata> obj ) {
	ltkt_metadata * md = obj.get_version()->get_metadata();
	md->del_writer();
	md->wakeup( fr->get_graph() );
    }
};

//----------------------------------------------------------------------
// Tags class for each dependency type
//----------------------------------------------------------------------
// Whole-function dependency tags
class function_tags : public function_tags_base { };

// Input dependency tags
class indep_tags : public indep_tags_base, public list_tags {
    template<typename MetaData, typename Task, template<typename T> class DepTy>
    friend class dep_traits;
/*
    ltkt_metadata::tag_t wr_tag;
#if OBJECT_COMMUTATIVITY
    ltkt_metadata::tag_t c_tag;
#endif
#if OBJECT_REDUCTION
    ltkt_metadata::tag_t r_tag;
#endif
*/
};

// Output dependency tags require fully serialized tags in the worst case
class outdep_tags : public outdep_tags_base, public serial_dep_tags {
    template<typename MetaData, typename Task, template<typename T> class DepTy>
    friend class dep_traits;
};

// Input/output dependency tags require fully serialized tags
class inoutdep_tags : public inoutdep_tags_base, public serial_dep_tags {
    template<typename MetaData, typename Task, template<typename T> class DepTy>
    friend class dep_traits;
};

// Commutative input/output dependency tags
#if OBJECT_COMMUTATIVITY
class cinoutdep_tags : public cinoutdep_tags_base, public list_tags {
    template<typename MetaData, typename Task, template<typename T> class DepTy>
    friend class dep_traits;
/*
    ltkt_metadata::tag_t wr_tag;
    ltkt_metadata::tag_t rd_tag;
#if OBJECT_REDUCTION
    ltkt_metadata::tag_t r_tag;
#endif
*/
};
#endif

// Reduction dependency tags
#if OBJECT_REDUCTION
class reduction_tags : public reduction_tags_base<ltkt_metadata>, public list_tags {
    template<typename MetaData, typename Task, template<typename T> class DepTy>
    friend class dep_traits;
/*
    ltkt_metadata::tag_t wr_tag;
    ltkt_metadata::tag_t rd_tag;
#if OBJECT_COMMUTATIVITY
    ltkt_metadata::tag_t c_tag;
#endif
*/
};
#endif

// Popdep (input) dependency tags - fully serialized with other pop and pushpop
class popdep_tags : public popdep_tags_base<ltkt_metadata>, public list_tags {
    template<typename MetaData, typename Task, template<typename T> class DepTy>
    friend class dep_traits;
    // ltkt_metadata::tag_t rd_tag;

public:
    popdep_tags( queue_version<ltkt_metadata> * parent )
	: popdep_tags_base( parent ) { }
};

// Pushpopdep (input/output) dependency tags - fully serialized with other
// pop and pushpop
class pushpopdep_tags : public pushpopdep_tags_base<ltkt_metadata>, public list_tags {
    template<typename MetaData, typename Task, template<typename T> class DepTy>
    friend class dep_traits;
    // ltkt_metadata::tag_t rd_tag;

public:
    pushpopdep_tags( queue_version<ltkt_metadata> * parent )
	: pushpopdep_tags_base( parent ) { }
};

// Pushdep (output) dependency tags
class pushdep_tags : public pushdep_tags_base<ltkt_metadata> {
    template<typename MetaData, typename Task, template<typename T> class DepTy>
    friend class dep_traits;

public:
    pushdep_tags( queue_version<ltkt_metadata> * parent )
	: pushdep_tags_base( parent ) { }
};


//----------------------------------------------------------------------
// Dependency handling traits to track task-object dependencies
//----------------------------------------------------------------------
// indep traits for objects
template<>
struct dep_traits<ltkt_metadata, task_metadata, indep> {
    template<typename T>
    static void arg_issue( task_metadata * fr, indep<T> & obj_ext,
			   typename indep<T>::dep_tags * tags ) {
	ltkt_metadata * md = obj_ext.get_version()->get_metadata();
	tags->copy_tags( md );
	md->add_reader();
    }
    template<typename T>
    static
    bool arg_ready( indep<T> & obj_int, typename indep<T>::dep_tags & tags ) {
	ltkt_metadata * md = obj_int.get_version()->get_metadata();
	bool r = md->chk_writer_tag( tags.wr_tag )
#if OBJECT_COMMUTATIVITY
	    & md->chk_commutative_tag( tags.c_tag )
#endif
#if OBJECT_REDUCTION
	    & md->chk_reduction_tag( tags.r_tag )
#endif
	    ;
	// errs() << "ready indep " << &tags << "? " << r << "\n";
	return r;
    }
    template<typename T>
    static
    bool arg_ini_ready( const indep<T> & obj_ext ) {
	const ltkt_metadata * md = obj_ext.get_version()->get_metadata();
	return !md->has_writers()
#if OBJECT_COMMUTATIVITY
	    & !md->has_commutative()
#endif
#if OBJECT_REDUCTION
	    & !md->has_reductions()
#endif
	    ;
    }
    template<typename T>
    static
    void arg_release( task_metadata * fr, indep<T> & obj,
		      typename indep<T>::dep_tags & tags ) {
	ltkt_metadata * md = obj.get_version()->get_metadata();
	md->del_reader();
	md->wakeup( fr->get_graph() );
    }
};

// indep traits for tokens
template<>
struct dep_traits<token_metadata, task_metadata, indep> {
    template<typename T>
    static void arg_issue( task_metadata * fr, indep<T> & obj_ext,
			   typename indep<T>::dep_tags * tags ) {
	token_metadata * md = obj_ext.get_version()->get_metadata();
	tags->copy_tags( md );
	md->add_reader();
    }
    template<typename T>
    static
    bool arg_ready( indep<T> & obj_int, typename indep<T>::dep_tags & tags ) {
	token_metadata * md = obj_int.get_version()->get_metadata();
	return md->chk_writer_tag( tags.wr_tag );
    }
    template<typename T>
    static
    bool arg_ini_ready( const indep<T> & obj_ext ) {
	const token_metadata * md = obj_ext.get_version()->get_metadata();
	return !md->has_writers();
    }
    template<typename T>
    static
    void arg_release( task_metadata * fr, indep<T> & obj,
		      typename indep<T>::dep_tags & tags ) {
	ltkt_metadata * md = obj.get_version()->get_metadata();
	md->del_reader();
	md->wakeup( fr->get_graph() );
    }
};


// output dependency traits for objects
template<>
struct dep_traits<ltkt_metadata, task_metadata, outdep> {
    template<typename T>
    static
    void arg_issue( task_metadata * fr, outdep<T> & obj_ext,
		    typename outdep<T>::dep_tags * tags ) {
	assert( obj_ext.get_version()->is_versionable() ); // enforced by applicators
	obj_ext.get_version()->get_metadata()->add_writer();
    }
    template<typename T>
    static
    bool arg_ready( outdep<T> & obj_ext, typename outdep<T>::dep_tags & tags ) {
	assert( obj_ext.get_version()->is_versionable() ); // enforced by applicators
	return true;
    }
    template<typename T>
    static
    bool arg_ini_ready( const outdep<T> & obj ) {
	assert( obj.get_version()->is_versionable() ); // enforced by applicators
	return true;
    }
    template<typename T>
    static
    void arg_release( task_metadata * fr, outdep<T> & obj,
		      typename outdep<T>::dep_tags & tags ) {
	serial_dep_traits::arg_release( fr, obj );
    }
};

// inout dependency traits for objects
template<>
struct dep_traits<ltkt_metadata, task_metadata, inoutdep> {
    template<typename T>
    static
    void arg_issue( task_metadata * fr, inoutdep<T> & obj_ext,
		    typename inoutdep<T>::dep_tags * tags ) {
	serial_dep_traits::arg_issue( fr, obj_ext, tags );
    }
    template<typename T>
    static
    bool arg_ready( inoutdep<T> & obj_ext,
		    typename inoutdep<T>::dep_tags & tags ) {
	return serial_dep_traits::arg_ready( obj_ext, tags );
    }
    template<typename T>
    static
    bool arg_ini_ready( const inoutdep<T> & obj_ext ) {
	return serial_dep_traits::arg_ini_ready( obj_ext );
    }
    template<typename T>
    static
    void arg_release( task_metadata * fr, inoutdep<T> & obj,
		      typename inoutdep<T>::dep_tags & tags ) {
	serial_dep_traits::arg_release( fr, obj );
    }
};

// inout dependency traits for tokens
template<>
struct dep_traits<token_metadata, task_metadata, inoutdep> {
    template<typename T>
    static
    void arg_issue( task_metadata * fr, inoutdep<T> & obj_ext,
		    typename inoutdep<T>::dep_tags * tags ) {
	token_metadata * md = obj_ext.get_version()->get_metadata();
	tags->rd_tag  = md->get_reader_tag();
	tags->wr_tag  = md->get_writer_tag();
	md->add_writer();
	md->update_depth( fr->get_depth() );
    }
    template<typename T>
    static
    bool arg_ready( inoutdep<T> & obj_ext,
		    typename inoutdep<T>::dep_tags & tags ) {
	token_metadata * md = obj_ext.get_version()->get_metadata();
	return md->chk_reader_tag( tags.rd_tag )
	    & md->chk_writer_tag( tags.wr_tag );
    }
    template<typename T>
    static
    bool arg_ini_ready( const inoutdep<T> & obj_ext ) {
	const token_metadata * md = obj_ext.get_version()->get_metadata();
	return !md->has_readers() & !md->has_writers();
    }
    template<typename T>
    static
    void arg_release( task_metadata * fr, inoutdep<T> & obj,
		      typename inoutdep<T>::dep_tags & tags ) {
	token_metadata * md = obj.get_version()->get_metadata();
	md->del_writer();
    }
};

// cinout dependency traits
#if OBJECT_COMMUTATIVITY
template<>
struct dep_traits<ltkt_metadata, task_metadata, cinoutdep> {
    template<typename T>
    static
    void arg_issue( task_metadata * fr, cinoutdep<T> & obj_ext,
		    typename cinoutdep<T>::dep_tags * tags ) {
	ltkt_metadata * md = obj_ext.get_version()->get_metadata();
	tags->copy_tags( md );
	md->add_commutative();
	// TODO: Parallel cinoutdeps have different depths!
	md->update_depth( fr->get_depth() );
    }
    template<typename T>
    static
    bool arg_ready( cinoutdep<T> & obj,
		    typename cinoutdep<T>::dep_tags & tags ) {
	ltkt_metadata * md = obj.get_version()->get_metadata();
	bool r = false;
	if( md->chk_reader_tag( tags.rd_tag )
	    & md->chk_writer_tag( tags.wr_tag )
#if OBJECT_REDUCTION
	    & md->chk_reduction_tag( tags.r_tag )
#endif
	    )
	    r = md->commutative_try_acquire();
	// errs() << "ready cinoutdep " << &tags << "? " << r << "\n";
	return r;
    }
    template<typename T>
    static
    bool arg_ini_ready( cinoutdep<T> obj_ext ) {
	ltkt_metadata * md = obj_ext.get_version()->get_metadata();
	if( !md->has_readers() & !md->has_writers()
#if OBJECT_REDUCTION
	    & !md->has_reductions()
#endif
	    )
	    return md->commutative_try_acquire();
	return false;
    }
    template<typename T>
    static
    void arg_ini_ready_undo( cinoutdep<T> obj_ext ) {
	obj_ext.get_version()->get_metadata()->commutative_release();
    }
    template<typename T>
    static
    void arg_release( task_metadata * fr, cinoutdep<T> obj,
		      typename cinoutdep<T>::dep_tags & tags ) {
	ltkt_metadata * md = obj.get_version()->get_metadata();
	md->del_commutative();
	md->commutative_release();
	md->wakeup( fr->get_graph() );
    }
};
#endif

// reduction dependency traits
#if OBJECT_REDUCTION
template<>
struct dep_traits<ltkt_metadata, task_metadata, reduction> {
    template<typename T>
    static
    void arg_issue( task_metadata * fr, reduction<T> & obj_ext,
		    typename reduction<T>::dep_tags * tags ) {
	ltkt_metadata * md = obj_ext.get_version()->get_metadata();
	tags->copy_tags( md );
	md->add_reduction();
	// TODO: Parallel reductions have different depths!
	md->update_depth( fr->get_depth() );
    }
    template<typename T>
    static
    bool arg_ready( reduction<T> & obj_int,
		    typename reduction<T>::dep_tags & tags ) {
	ltkt_metadata * md = tags.ext_version->get_metadata();
	bool r = md->chk_reader_tag( tags.rd_tag )
	    & md->chk_writer_tag( tags.wr_tag )
#if OBJECT_COMMUTATIVITY
	    & md->chk_commutative_tag( tags.c_tag )
#endif
	    ;
	// errs() << "ready reduction " << &tags << "? " << r << "\n";
	return r;
    }
    template<typename T>
    static
    bool arg_ini_ready( reduction<T> obj_ext ) {
	const ltkt_metadata * md = obj_ext.get_version()->get_metadata();
	return !md->has_readers() & !md->has_writers()
#if OBJECT_COMMUTATIVITY
	    & !md->has_commutative()
#endif
	    ;
    }
    template<typename T>
    static
    void arg_release( task_metadata * fr, reduction<T> obj_int,
		      typename reduction<T>::dep_tags & tags ) {
	ltkt_metadata * md = tags.ext_version->get_metadata();
	md->del_reduction();
	md->wakeup( fr->get_graph() );
    }
};
#endif

// popdep traits for queues -- TODO: use token_metadata instead!
template<>
struct dep_traits<ltkt_metadata, task_metadata, popdep> {
    template<typename T>
    static void arg_issue( task_metadata * fr, popdep<T> & obj_int,
			   typename popdep<T>::dep_tags * tags ) {
	ltkt_metadata * md = obj_int.get_version()->get_metadata();
	tags->copy_tags( md );
	md->add_reader();
    }
    template<typename T>
    static
    bool arg_ready( popdep<T> & obj_int, typename popdep<T>::dep_tags & tags ) {
	ltkt_metadata * md = obj_int.get_version()->
	    get_parent()->get_metadata();
	return md->chk_reader_tag( tags.rd_tag );
    }
    template<typename T>
    static
    bool arg_ini_ready( const popdep<T> & obj_ext ) {
	const ltkt_metadata * md = obj_ext.get_version()->get_metadata();
	return !md->has_readers();
    }
    template<typename T>
    static
    void arg_release( task_metadata * fr, popdep<T> & obj_int,
		      typename popdep<T>::dep_tags & tags ) {
	ltkt_metadata * md = tags.ext_version->get_metadata();
	obj_int.get_version()->get_metadata()->del_reader();
	md->wakeup( fr->get_graph() );
    }
};

// pushdep dependency traits for queues
template<>
struct dep_traits<ltkt_metadata, task_metadata, pushdep> {
    template<typename T>
    static
    void arg_issue( task_metadata * fr, pushdep<T> & obj_int,
		    typename pushdep<T>::dep_tags * tags ) {
    }
    template<typename T>
    static
    bool arg_ready( pushdep<T> & obj_int,
		    typename pushdep<T>::dep_tags & tags ) {
	return true;
    }
    template<typename T>
    static
    bool arg_ini_ready( const pushdep<T> & obj_ext ) {
	return true;
    }
    template<typename T>
    static
    void arg_release( task_metadata * fr, pushdep<T> & obj_int,
		      typename pushdep<T>::dep_tags & tags ) {
    }
};

typedef ltkt_metadata obj_metadata;
typedef ltkt_metadata queue_metadata; // could also be token_metadata!

} // end of namespace obj

#endif // LTICKETS_H
